function p = logpot(z,m,ze,smooth,xopt,cutoff,cont)
%LOGPOT Evaluation of logarithmic potential.
%   P = LOGPOT(Z,M,ZE,SMOOTH,XOPT,CUTOFF,CONT) is a reference
%   implementation of the logarithmic potential -M*LOG(ABS(Z)) with 
%   masses M at coordinates Z evaluated at ZE using the smoother SMOOTH =
%   {'dirac'} | 'rankine' | 'scully' | 'oseen' with parameters XOPT
%   (default 1e-3), CUTOFF (default 2e-3) and CONT {'on'} | 'off'.
%
%   XOPT is the distance from the potential to the extreme point of
%   abs(P1), where P1 is the corresponding smoothed Harmonic potential
%   as computed by INVPOT. CUTOFF is the cut-off radius and it is
%   required that XOPT < CUTOFF. If CONT is 'on', then the smoother is
%   scaled to ensure continuity at the cut-off.
%
%   Example:
%     z = 0; m = 1; % a single potential at z = 0
%     x = linspace(-3e-3,3e-3);
%     [X,Y] = meshgrid(x,x);
%     ze = complex(X,Y); zf = complex(x,x);
%     cc = 5:0.25:7;
%     s = {'dirac' 'rankine' 'scully' 'oseen'};
%     for i = 1:size(s,2)
%       p{i} = logpot(z,m,ze,s{i});
%       figure(1), subplot(2,2,i);
%       contour(x,x,abs(p{i}),cc); axis square, title(s{i});
%       q{i} = logpot(z,m,zf,s{i});
%     end
%     figure(2);
%     xx = sign(x).*abs(zf);
%     plot(xx,abs(q{1}),'b',xx,abs(q{2}),'r', ...
%          xx,abs(q{3}),'g',xx,abs(q{4}),'k');
%     legend(s{:}); axis([xx([1 end]) 5 10]);
%
%   See also FMM2D, INVPOT.

% S. Engblom 2007-05-16 (Revision)
% P. Deglaire 2006-07-17 (Modification)
% S. Engblom 2006-07-14

% defaults
if nargin < 7
  cont = 'on';
  if nargin < 6
    cutoff = 2e-3;
    if nargin < 5
      xopt = 1e-3;
      if nargin < 4
        smooth = 'dirac';
        if nargin < 3
          ze = z;
        end
      end
    end
  end
end

if xopt >= cutoff, error('Illegal parameters.'); end

if isempty(ze), ze = z; end

N = numel(z);
NE = numel(ze);
p = zeros(size(ze));

switch smooth
 case 'dirac'
  for i = 1:N
    for j = 1:NE
      if ze(j) ~= z(i) % protect from exactly zero
	p(j) = p(j)-m(i)*reallog(abs(ze(j)-z(i)));
      end
    end
  end

 case 'rankine'
  cutoff = xopt;
  for i = 1:N
    for j = 1:NE
      d = ze(j)-z(i);
      ad = abs(d);
      if ad >= cutoff
	p(j) = p(j)-m(i)*reallog(ad);
      else
	p(j) = p(j)-m(i)*(((ad/cutoff)^2-1)/2+reallog(cutoff));
      end
    end
  end

 case 'scully'
  shape = xopt;
  if strcmp(cont,'on')
    scale = reallog(cutoff^2)/reallog(shape^2+cutoff^2);
  else
    scale = 1;
  end
  for i = 1:N
    for j = 1:NE
      d = ze(j)-z(i);
      ad = abs(d);
      if ad >= cutoff
	p(j) = p(j)-m(i)*reallog(ad);
      else
	p(j) = p(j)-m(i)*(reallog(shape^2+abs(d)^2)/2*scale);
      end
    end
  end

 case 'oseen'
  shape = (1.1209064227785340320*cutoff/xopt)^2;
  if strcmp(cont,'on')
    scale = reallog(cutoff)/(reallog(cutoff)+0.5*expint(shape));
  else
    scale = 1;
  end
  for i = 1:N
    for j = 1:NE
      d = ze(j)-z(i);
      ad = abs(d);
      if ad >= cutoff
        p(j) = p(j)-m(i)*reallog(ad);
      elseif d ~= 0 % protect from exactly zero
        p(j) = p(j)-m(i)*((reallog(ad)+ ...
			   0.5*expint(shape*(ad/cutoff)^2))*scale);
      else
        p(j) = p(j)+m(i)*0.5*(reallog(shape/cutoff^2)-psi(1))*scale;
      end
    end
  end

 otherwise, error('Unknown smoother.');
end
